/*
 * userial -- a USB to I2C converter
 *
 * Copyright (c) 2008-2009, Thomas Pircher <tehpeh@gmx.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *****************************************************************************/
#include "hardware.h"
#include "i2c.h"
#include "cbuf.h"
#include "spi.h"
#include "command.h"

#include "cdc.h"
#include "cdc_descriptors.h"
#include <LUFA/Version.h>
//#include <LUFA/Common/ButtLoadTag.h>
#include <LUFA/Drivers/USB/USB.h>
#include <LUFA/Scheduler/Scheduler.h>

// Project Tags, for reading out using the ButtLoad project
BUTTLOADTAG(ProjName,    "userial USB/I2C/SPI");
BUTTLOADTAG(BuildTime,   __TIME__);
BUTTLOADTAG(BuildDate,   __DATE__);
BUTTLOADTAG(LUFAVersion, "userial v" VERSION_STRING);

// Scheduler Task List
TASK_LIST
{
    { .Task = USB_USBTask,  .TaskStatus = TASK_STOP },
    { .Task = CDC_Task,     .TaskStatus = TASK_STOP },
};

/**
 * Contains the current baud rate and other settings of the virtual serial
 * port. While this demo does not use the physical USART and thus does not use
 * these settings, they must still be retained and returned to the host upon
 * request or the host will assume the device is non-functional.
 *
 * These values are set by the host via a class-specific request, however they
 * are not required to be used accurately.  It is possible to completely ignore
 * these value or use other settings as the host is completely unaware of the
 * physical serial link characteristics and instead sends and recieves data in
 * endpoint streams.
 *****************************************************************************/
CDC_Line_Coding_t LineCoding = { .BaudRateBPS = 9600,
                                 .CharFormat =  OneStopBit,
                                 .ParityType =  Parity_None,
                                 .DataBits =    8};

/**
 * The receiver buffer
 *****************************************************************************/
cbf_t rx_buf;

/**
 * The transmission buffer
 *****************************************************************************/
cbf_t tx_buf;

/**
 * The command buffer
 *****************************************************************************/
cbf_t cmd_buf;


/**
 * The main function.
 *
 * \return This function does never return.
 *****************************************************************************/
int main(void)
{
    hw_init();
    i2c_init();
    spi_init();
    cbf_init(&rx_buf);
    cbf_init(&tx_buf);

    // Indicate USB not ready
    UpdateStatus(Status_USBNotReady);

    // Initialize Scheduler so that it can be used
    Scheduler_Init();

    // Initialize USB Subsystem
    USB_Init();

    // Scheduling - routine never returns, so put this last in the main function
    Scheduler_Start();
}

/**
 * Event handler for the USB_Connect event. This indicates that the device is
 * enumerating via the status LEDs and starts the library USB task to begin the
 * enumeration and USB management process.
 *****************************************************************************/
EVENT_HANDLER(USB_Connect)
{
    // Start USB management task
    Scheduler_SetTaskMode(USB_USBTask, TASK_RUN);

    // Indicate USB enumerating
    UpdateStatus(Status_USBEnumerating);
}

/**
 * Event handler for the USB_Disconnect event. This indicates that the device
 * is no longer connected to a host via the status LEDs and stops the USB
 * management and CDC management tasks.
 *****************************************************************************/
EVENT_HANDLER(USB_Disconnect)
{
    // Stop running CDC and USB management tasks
    Scheduler_SetTaskMode(CDC_Task, TASK_STOP);
    Scheduler_SetTaskMode(USB_USBTask, TASK_STOP);

    // Indicate USB not ready
    UpdateStatus(Status_USBNotReady);
}

/**
 * Event handler for the USB_ConfigurationChanged event. This is fired when the
 * host set the current configuration of the USB device after enumeration - the
 * device endpoints are configured and the CDC management task started.
 *****************************************************************************/
EVENT_HANDLER(USB_ConfigurationChanged)
{
    // Setup CDC Notification, Rx and Tx Endpoints
    Endpoint_ConfigureEndpoint(CDC_NOTIFICATION_EPNUM, EP_TYPE_INTERRUPT,
            ENDPOINT_DIR_IN, CDC_NOTIFICATION_EPSIZE,
            ENDPOINT_BANK_SINGLE);

    Endpoint_ConfigureEndpoint(CDC_TX_EPNUM, EP_TYPE_BULK,
            ENDPOINT_DIR_IN, CDC_TXRX_EPSIZE,
            ENDPOINT_BANK_SINGLE);

    Endpoint_ConfigureEndpoint(CDC_RX_EPNUM, EP_TYPE_BULK,
            ENDPOINT_DIR_OUT, CDC_TXRX_EPSIZE,
            ENDPOINT_BANK_SINGLE);

    // Indicate USB connected and ready
    UpdateStatus(Status_USBReady);

    // Start CDC task
    Scheduler_SetTaskMode(CDC_Task, TASK_RUN);
}

/**
 * Event handler for the USB_UnhandledControlPacket event. This is used to
 * catch standard and class specific control requests that are not handled
 * internally by the USB library (including the CDC control commands, which are
 * all issued via the control endpoint), so that they can be handled
 * appropriately for the application.
 *****************************************************************************/
EVENT_HANDLER(USB_UnhandledControlPacket)
{
    uint8_t* LineCodingData = (uint8_t*)&LineCoding;

    // Process CDC specific control requests
    switch (bRequest)
    {
        case REQ_GetLineEncoding:
            if (bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
            {
                // Acknowedge the SETUP packet, ready for data transfer
                Endpoint_ClearSetupReceived();

                // Write the line coding data to the control endpoint
                Endpoint_Write_Control_Stream_LE(LineCodingData, sizeof(CDC_Line_Coding_t));

                // Finalize the stream transfer to send the last packet or clear the host abort
                Endpoint_ClearSetupOUT();
            }

            break;
        case REQ_SetLineEncoding:
            if (bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
            {
                // Acknowedge the SETUP packet, ready for data transfer
                Endpoint_ClearSetupReceived();

                // Read the line coding data in from the host into the global struct
                Endpoint_Read_Control_Stream_LE(LineCodingData, sizeof(CDC_Line_Coding_t));

                // Finalize the stream transfer to clear the last packet from the host
                Endpoint_ClearSetupIN();
            }

            break;
        case REQ_SetControlLineState:
            if (bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
            {
#if 0
                /* NOTE: Here you can read in the line state mask from the host, to get the current state of the output handshake
                         lines. The mask is read in from the wValue parameter, and can be masked against the CONTROL_LINE_OUT_* masks
                         to determine the RTS and DTR line states using the following code:
                */

                uint16_t wIndex = Endpoint_Read_Word_LE();

                // Do something with the given line states in wIndex
#endif

                // Acknowedge the SETUP packet, ready for data transfer
                Endpoint_ClearSetupReceived();

                // Send an empty packet to acknowedge the command
                Endpoint_ClearSetupIN();
            }

            break;
    }
}

/**
 * Function to manage status updates to the user. This is done via LEDs on the
 * given board, if available, but may be changed to log to a serial port, or
 * anything else that is suitable for status updates.
 *
 * \param CurrentStatus  Current status of the system, from the CDC_StatusCodes_t enum
 *****************************************************************************/
void UpdateStatus(uint8_t CurrentStatus)
{
  // Set the LED mask to the appropriate LED mask based on the given status code
  switch (CurrentStatus)
  {
      case Status_USBNotReady:
          led_off(HW_LED1);
          led_on(HW_LED1);
          break;
      case Status_USBEnumerating:
          led_toggle(HW_LED1 | HW_LED2);
          break;
      case Status_USBReady:
          led_off(HW_LED1 | HW_LED2);
          break;
  }
}

/**
 * The CDC task.
 * Function to manage CDC data transmission and reception to and from the host.
 *****************************************************************************/
TASK(CDC_Task)
{
    bool endpoint_full;

    // Select the Serial Rx Endpoint
    Endpoint_SelectEndpoint(CDC_RX_EPNUM);
    if (Endpoint_ReadWriteAllowed()) {
        led_on(HW_LED1);
        while (Endpoint_BytesInEndpoint()) {
            uint8_t c = Endpoint_Read_Byte();
            cbf_put(&rx_buf, c);
        }
        Endpoint_ClearCurrentBank();
        led_off(HW_LED1);
    }

    led_on(HW_LED2);
    while (cmd_scan(&rx_buf, &tx_buf)) {
        if (cmd_parse(&rx_buf, &cmd_buf, &tx_buf)) {
            if (!cbf_isempty(&cmd_buf)) {
                cmd_exec(&cmd_buf, &tx_buf);
            }
        }
    }
    led_off(HW_LED2);

    // Select the Serial Tx Endpoint
    Endpoint_SelectEndpoint(CDC_TX_EPNUM);
    if (!cbf_isempty(&tx_buf)) {
        led_on(HW_LED1);
        while (!(Endpoint_ReadWriteAllowed()));

        while (!cbf_isempty(&tx_buf) && (Endpoint_BytesInEndpoint() < CDC_TXRX_EPSIZE)) {
            uint8_t c = cbf_get(&tx_buf);
            Endpoint_Write_Byte(c);
        }
        endpoint_full = Endpoint_BytesInEndpoint() == CDC_TXRX_EPSIZE;
        Endpoint_ClearCurrentBank();

        // If the endpoint is full, send an empty package to flush the output queue.
        if (endpoint_full) {
            while (Endpoint_BytesInEndpoint());
            Endpoint_ClearCurrentBank();
        }
        led_off(HW_LED1);
    }
}
